@@ -72,9 +72,11 @@ class WindowPreview extends Tooltips.TooltipBase {
     constructor(item, metaWindow, previewScale, showLabel) {
         super(item.actor);
         this._applet = item._applet;
-        this.uiScale = St.ThemeContext.get_for_stage(global.stage).scale_factor;
+        this.metaWindow = metaWindow;
+        let x = this.windowActor;
+        this.uiScale = global.ui_scale;
         this.thumbScale = previewScale;
-        this.muffinWindow = metaWindow.get_compositor_private();
+
         this._sizeChangedId = null;
         this.thumbnail = null;
 
@@ -96,6 +98,17 @@ class WindowPreview extends Tooltips.TooltipBase {
         this.actor.add_actor(this.thumbnailBin);
     }
 
+    get windowActor() {
+        let actor = this.metaWindow.get_compositor_private();
+
+        if (actor) {
+            return actor;
+        } else {
+            log("metaWindow has no actor!");
+            return null;
+        }
+    }
+
     _onEnterEvent(actor, event) {
         if (this._applet._tooltipShowing)
             this.show();
@@ -122,22 +135,28 @@ class WindowPreview extends Tooltips.TooltipBase {
         if (!this.actor || this._applet._menuOpen)
             return;
 
-        let windowTexture = this.muffinWindow.get_texture();
-        let [width, height] = this._getScaledTextureSize(windowTexture);
-
         if (this.thumbnail) {
             this.thumbnailBin.set_child(null);
             this.thumbnail.destroy();
             this.thumbnail = null;
         }
 
+        let windowTexture = this.windowActor.get_texture();
+
+        if (!windowTexture) {
+            this.actor.hide();
+            return;
+        }
+
+        let [width, height] = this._getScaledTextureSize(windowTexture);
+
         this.thumbnail = new Clutter.Clone({
             source: windowTexture,
             width: width,
             height: height
         });
 
-        this._sizeChangedId = this.muffinWindow.connect('size-changed', () => {
+        this._sizeChangedId = this.windowActor.connect('size-changed', () => {
             let [width, height] = this._getScaledTextureSize(windowTexture);
             this.thumbnail.set_size(width, height);
             this._set_position();
@@ -155,7 +174,7 @@ class WindowPreview extends Tooltips.TooltipBase {
 
     hide() {
         if (this._sizeChangedId != null) {
-            this.muffinWindow.disconnect(this._sizeChangedId);
+            this.windowActor.disconnect(this._sizeChangedId);
             this._sizeChangedId = null;
         }
         if (this.thumbnail) {
@@ -212,7 +231,7 @@ class WindowPreview extends Tooltips.TooltipBase {
 
     _destroy() {
         if (this._sizeChangedId != null) {
-            this.muffinWindow.disconnect(this._sizeChangedId);
+            this.windowActor.disconnect(this._sizeChangedId);
             this.sizeChangedId = null;
         }
         if (this.thumbnail) {
@@ -224,7 +243,6 @@ class WindowPreview extends Tooltips.TooltipBase {
             this.actor.destroy();
             this.actor = null;
         }
-        this.muffinWindow = null;
     }
 }
 
@@ -241,7 +259,7 @@ class AppMenuButton {
         this.metaWindow = metaWindow;
         this.alert = alert;
         this.labelVisible = false;
-        this.window_signals = new SignalManager.SignalManager(null);
+        this._signals = new SignalManager.SignalManager();
 
         if (this._applet.orientation == St.Side.TOP)
             this.actor.add_style_class_name('top');
@@ -253,23 +271,23 @@ class AppMenuButton {
             this.actor.add_style_class_name('right');
 
         this.actor._delegate = this;
-        this.actor.connect('button-release-event', Lang.bind(this, this._onButtonRelease));
-        this.actor.connect('button-press-event', Lang.bind(this, this._onButtonPress));
+        this._signals.connect(this.actor, 'button-release-event', Lang.bind(this, this._onButtonRelease));
+        this._signals.connect(this.actor, 'button-press-event', Lang.bind(this, this._onButtonPress));
 
-        this.actor.connect('get-preferred-width',
+        this._signals.connect(this.actor, 'get-preferred-width',
                 Lang.bind(this, this._getPreferredWidth));
-        this.actor.connect('get-preferred-height',
+        this._signals.connect(this.actor, 'get-preferred-height',
                 Lang.bind(this, this._getPreferredHeight));
-        this.actor.connect('allocate', Lang.bind(this, this._allocate));
+        this._signals.connect(this.actor, 'allocate', Lang.bind(this, this._allocate));
 
         this.progressOverlay = new St.Widget({ style_class: "progress", reactive: false, important: true  });
 
         this.actor.add_actor(this.progressOverlay);
 
         this._iconBox = new Cinnamon.Slicer({ name: 'appMenuIcon' });
-        this._iconBox.connect('style-changed',
+        this._signals.connect(this._iconBox, 'style-changed',
                               Lang.bind(this, this._onIconBoxStyleChanged));
-        this._iconBox.connect('notify::allocation',
+        this._signals.connect(this._iconBox, 'notify::allocation',
                               Lang.bind(this, this._updateIconBoxClipAndGeometry));
         this.actor.add_actor(this._iconBox);
 
@@ -316,13 +334,13 @@ class AppMenuButton {
             this._menuManager.addMenu(this.rightClickMenu);
 
             this._draggable = DND.makeDraggable(this.actor, null, this._applet.actor);
-            this._draggable.connect('drag-begin', Lang.bind(this, this._onDragBegin));
-            this._draggable.connect('drag-cancelled', Lang.bind(this, this._onDragCancelled));
-            this._draggable.connect('drag-end', Lang.bind(this, this._onDragEnd));
+            this._signals.connect(this._draggable, 'drag-begin', Lang.bind(this, this._onDragBegin));
+            this._signals.connect(this._draggable, 'drag-cancelled', Lang.bind(this, this._onDragCancelled));
+            this._signals.connect(this._draggable, 'drag-end', Lang.bind(this, this._onDragEnd));
         }
 
         this.onPanelEditModeChanged();
-        global.settings.connect('changed::panel-edit-mode', Lang.bind(this, this.onPanelEditModeChanged));
+        this._signals.connect(global.settings, 'changed::panel-edit-mode', Lang.bind(this, this.onPanelEditModeChanged));
 
         this._windows = this._applet._windows;
 
@@ -337,12 +355,12 @@ class AppMenuButton {
         if (this.alert)
             this.getAttention();
 
-        this.window_signals.connect(this.metaWindow, 'notify::title', this.setDisplayTitle, this);
-        this.window_signals.connect(this.metaWindow, "notify::minimized", this.setDisplayTitle, this);
-        this.window_signals.connect(this.metaWindow, "notify::tile-type", this.setDisplayTitle, this);
-        this.window_signals.connect(this.metaWindow, "notify::icon", this.setIcon, this);
-        this.window_signals.connect(this.metaWindow, "notify::appears-focused", this.onFocus, this);
-        this.window_signals.connect(this.metaWindow, "unmanaged", this.onUnmanaged, this);
+        this._signals.connect(this.metaWindow, 'notify::title', this.setDisplayTitle, this);
+        this._signals.connect(this.metaWindow, "notify::minimized", this.setDisplayTitle, this);
+        this._signals.connect(this.metaWindow, "notify::tile-type", this.setDisplayTitle, this);
+        this._signals.connect(this.metaWindow, "icon-changed", this.setIcon, this);
+        this._signals.connect(this.metaWindow, "notify::appears-focused", this.onFocus, this);
+        this._signals.connect(this.metaWindow, "unmanaged", this.onUnmanaged, this);
     }
 
     onUnmanaged() {
@@ -501,10 +519,11 @@ class AppMenuButton {
     }
 
     destroy() {
-        this.window_signals.disconnectAllSignals();
+        this._signals.disconnectAllSignals();
         this._tooltip.destroy();
-        if (this.rightClickMenu) {
+        if (!this.alert) {
             this.rightClickMenu.destroy();
+            this._menuManager.destroy();
         }
         this.actor.destroy();
     }
@@ -724,11 +743,6 @@ class AppMenuButton {
         let tracker = Cinnamon.WindowTracker.get_default();
         let app = tracker.get_window_app(this.metaWindow);
 
-        if (!app) {
-            setTimeout(() => this.setIcon(), 0);
-            return;
-        }
-
         this.icon_size = this._applet.icon_size;
 
         let icon = app ?
@@ -749,26 +763,29 @@ class AppMenuButton {
             return false;
 
         this._needsAttention = true;
-        let counter = 0;
-        this._flashButton(counter);
+        this._flashButton();
         return true;
     }
 
-    _flashButton(counter) {
+    _flashButton() {
         if (!this._needsAttention)
             return;
 
-        this.actor.add_style_class_name("window-list-item-demands-attention");
-        if (counter < 4) {
-            Mainloop.timeout_add(FLASH_INTERVAL, Lang.bind(this, function () {
-                if (this.actor.has_style_class_name("window-list-item-demands-attention")) {
-                    this.actor.remove_style_class_name("window-list-item-demands-attention");
-                }
-                Mainloop.timeout_add(FLASH_INTERVAL, Lang.bind(this, function () {
-                    this._flashButton(++counter);
-                }));
-            }));
-        }
+        let counter = 0;
+        let sc = "window-list-item-demands-attention";
+
+        Mainloop.timeout_add(FLASH_INTERVAL, () => {
+            if (!this._needsAttention)
+                return false;
+
+            if (this.actor.has_style_class_name(sc))
+                this.actor.remove_style_class_name(sc);
+            else
+                this.actor.add_style_class_name(sc);
+            let result = counter < 4;
+            counter++;
+            return result;
+        });
     }
 };
 
@@ -778,7 +795,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
 
         this._launcher = launcher;
         this._windows = launcher._applet._windows;
-        this.connect('open-state-changed', Lang.bind(this, this._onToggled));
+        this._signals.connect(this, 'open-state-changed', Lang.bind(this, this._onToggled));
 
         this.orientation = orientation;
         this.metaWindow = metaWindow;
@@ -794,7 +811,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
             Main.layoutManager.monitors.forEach(function (monitor, index) {
                 if (index === mw.get_monitor()) return;
                 item = new PopupMenu.PopupMenuItem(_("Move to the other monitor"));
-                item.connect('activate', function() {
+                this._signals.connect(item, 'activate', function() {
                     mw.move_to_monitor(index);
                 });
                 this.addMenuItem(item);
@@ -804,7 +821,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
             Main.layoutManager.monitors.forEach(function (monitor, index) {
                 if (index === mw.get_monitor()) return;
                 item = new PopupMenu.PopupMenuItem(_("Move to monitor %d").format(index + 1));
-                item.connect('activate', function() {
+                this._signals.connect(item, 'activate', function() {
                     mw.move_to_monitor(index);
                 });
                 this.addMenuItem(item);
@@ -815,13 +832,13 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         if ((length = global.screen.n_workspaces) > 1) {
             if (mw.is_on_all_workspaces()) {
                 item = new PopupMenu.PopupMenuItem(_("Only on this workspace"));
-                item.connect('activate', function() {
+                this._signals.connect(item, 'activate', function() {
                     mw.unstick();
                 });
                 this.addMenuItem(item);
             } else {
                 item = new PopupMenu.PopupMenuItem(_("Visible on all workspaces"));
-                item.connect('activate', function() {
+                this._signals.connect(item, 'activate', function() {
                     mw.stick();
                 });
                 this.addMenuItem(item);
@@ -839,7 +856,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
                     if (i == curr_index)
                         ws.setSensitive(false);
 
-                    ws.connect('activate', function() {
+                    this._signals.connect(ws, 'activate', function() {
                         mw.change_workspace(global.screen.get_workspace_by_index(j));
                     });
                     item.menu.addMenuItem(ws);
@@ -853,24 +870,22 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         this.addMenuItem(subMenu);
 
         item = new PopupMenu.PopupIconMenuItem(_("About..."), "dialog-question", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this._launcher._applet, this._launcher._applet.openAbout));
+        this._signals.connect(item, 'activate', Lang.bind(this._launcher._applet, this._launcher._applet.openAbout));
         subMenu.menu.addMenuItem(item);
 
         item = new PopupMenu.PopupIconMenuItem(_("Configure..."), "system-run", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this._launcher._applet, this._launcher._applet.configureApplet));
+        this._signals.connect(item, 'activate', Lang.bind(this._launcher._applet, this._launcher._applet.configureApplet));
         subMenu.menu.addMenuItem(item);
 
         item = new PopupMenu.PopupIconMenuItem(_("Remove 'Window list'"), "edit-delete", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this, function() {
-            AppletManager._removeAppletFromPanel(this._launcher._applet._uuid, this._launcher._applet.instance_id);
-        }));
+        this._signals.connect(item, 'activate', (actor, event) => this._launcher._applet.confirmRemoveApplet(event));
         subMenu.menu.addMenuItem(item);
 
         this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
 
         // Close all/others
         item = new PopupMenu.PopupIconMenuItem(_("Close all"), "application-exit", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this, function() {
+        this._signals.connect(item, 'activate', Lang.bind(this, function() {
             for (let window of this._windows)
                 if (window.actor.visible &&
                    !window._needsAttention)
@@ -879,7 +894,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         this.addMenuItem(item);
 
         item = new PopupMenu.PopupIconMenuItem(_("Close others"), "window-close", St.IconType.SYMBOLIC);
-        item.connect('activate', Lang.bind(this, function() {
+        this._signals.connect(item, 'activate', Lang.bind(this, function() {
             for (let window of this._windows)
                 if (window.actor.visible &&
                     window.metaWindow != this.metaWindow &&
@@ -893,7 +908,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         // Miscellaneous
         if (mw.get_compositor_private().opacity != 255) {
             item = new PopupMenu.PopupMenuItem(_("Restore to full opacity"));
-            item.connect('activate', function() {
+            this._signals.connect(item, 'activate', function() {
                 mw.get_compositor_private().set_opacity(255);
             });
             this.addMenuItem(item);
@@ -901,12 +916,12 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
 
         if (mw.minimized) {
             item = new PopupMenu.PopupIconMenuItem(_("Restore"), "view-sort-descending", St.IconType.SYMBOLIC);
-            item.connect('activate', function() {
+            this._signals.connect(item, 'activate', function() {
                 Main.activateWindow(mw, global.get_current_time());
             });
         } else {
             item = new PopupMenu.PopupIconMenuItem(_("Minimize"), "view-sort-ascending", St.IconType.SYMBOLIC);
-            item.connect('activate', function() {
+            this._signals.connect(item, 'activate', function() {
                 mw.minimize();
             });
         }
@@ -914,19 +929,19 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
 
         if (mw.get_maximized()) {
             item = new PopupMenu.PopupIconMenuItem(_("Unmaximize"), "view-restore", St.IconType.SYMBOLIC);
-            item.connect('activate', function() {
+            this._signals.connect(item, 'activate', function() {
                 mw.unmaximize(Meta.MaximizeFlags.HORIZONTAL | Meta.MaximizeFlags.VERTICAL);
             });
         } else {
             item = new PopupMenu.PopupIconMenuItem(_("Maximize"), "view-fullscreen", St.IconType.SYMBOLIC);
-            item.connect('activate', function() {
+            this._signals.connect(item, 'activate', function() {
                 mw.maximize(Meta.MaximizeFlags.HORIZONTAL | Meta.MaximizeFlags.VERTICAL);
             });
         }
         this.addMenuItem(item);
 
         item = new PopupMenu.PopupIconMenuItem(_("Close"), "edit-delete", St.IconType.SYMBOLIC);
-        item.connect('activate', function() {
+        this._signals.connect(item, 'activate', function() {
             mw.delete(global.get_current_time());
         });
         this.addMenuItem(item);
@@ -998,14 +1013,15 @@ class CinnamonWindowListApplet extends Applet.Applet {
         this.settings.bind("window-preview-show-label", "showLabel", this._onPreviewChanged);
         this.settings.bind("window-preview-scale", "previewScale", this._onPreviewChanged);
 
-        this.signals.connect(global.screen, 'window-added', this._onWindowAdded, this);
+        this.signals.connect(global.screen, 'window-added', this._onWindowAddedAsync, this);
         this.signals.connect(global.screen, 'window-monitor-changed', this._onWindowMonitorChanged, this);
         this.signals.connect(global.screen, 'window-workspace-changed', this._onWindowWorkspaceChanged, this);
         this.signals.connect(global.screen, 'window-skip-taskbar-changed', this._onWindowSkipTaskbarChanged, this);
         this.signals.connect(global.screen, 'monitors-changed', this._updateWatchedMonitors, this);
         this.signals.connect(global.window_manager, 'switch-workspace', this._refreshAllItems, this);
+        this.signals.connect(Cinnamon.WindowTracker.get_default(), "window-app-changed", this._onWindowAppChanged, this);
 
-        this.actor.connect('style-changed', Lang.bind(this, this._updateSpacing));
+        this.signals.connect(this.actor, 'style-changed', Lang.bind(this, this._updateSpacing));
 
         global.settings.bind("panel-edit-mode", this.actor, "reactive", Gio.SettingsBindFlags.DEFAULT);
 
@@ -1020,6 +1036,7 @@ class CinnamonWindowListApplet extends Applet.Applet {
 
     on_applet_removed_from_panel() {
         this.signals.disconnectAllSignals();
+        this.settings.finalize();
     }
 
     on_applet_instances_changed() {
@@ -1095,6 +1112,10 @@ class CinnamonWindowListApplet extends Applet.Applet {
         this.manager.set_spacing(spacing * global.ui_scale);
     }
 
+    _onWindowAddedAsync(screen, metaWindow, monitor) {
+        Mainloop.timeout_add(20, Lang.bind(this, this._onWindowAdded, screen, metaWindow, monitor));
+    }
+
     _onWindowAdded(screen, metaWindow, monitor) {
         if (this._shouldAdd(metaWindow))
             this._addWindow(metaWindow, false);
@@ -1107,17 +1128,23 @@ class CinnamonWindowListApplet extends Applet.Applet {
             this._removeWindow(metaWindow);
     }
 
-    _onWindowWorkspaceChanged(screen, metaWindow, metaWorkspace) {
+    _refreshItemByMetaWindow(metaWindow) {
         let window = this._windows.find(win => (win.metaWindow == metaWindow));
 
         if (window)
             this._refreshItem(window);
     }
 
-    _onWindowSkipTaskbarChanged(screen, metaWindow) {
-        let window = this._windows.find(win => (win.metaWindow == metaWindow));
+    _onWindowWorkspaceChanged(screen, metaWindow, metaWorkspace) {
+        this._refreshItemByMetaWindow(metaWindow);
+    }
 
-        if (window && window.is_skip_taskbar()) {
+    _onWindowAppChanged(tracker, metaWindow) {
+        this._refreshItemByMetaWindow(metaWindow);
+    }
+
+    _onWindowSkipTaskbarChanged(screen, metaWindow) {
+        if (metaWindow && metaWindow.is_skip_taskbar()) {
             this._removeWindow(metaWindow);
             return;
         }
